#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "atomparser.h"

const char* AtomType[] = {"moov", "trak", "udta", "tref", "imap", "mdia",
    "minf", "stbl", "edts", "mdra", "rmra", "imag", "vnrp", "dinf"};

const char*  GenreList[] = {
    "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk",
    "Grunge", "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies",
    "Other", "Pop", "R&B", "Rap", "Reggae", "Rock",
    "Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks",
    "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk",
    "Fusion", "Trance", "Classical", "Instrumental", "Acid", "House",
    "Game", "Sound Clip", "Gospel", "Noise", "AlternRock", "Bass",
    "Soul", "Punk", "Space", "Meditative", "Instrumental Pop", "Instrumental Rock",
    "Ethnic", "Gothic", "Darkwave", "Techno-Industrial", "Electronic", "Pop-Folk",
    "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult", "Gangsta",
    "Top 40", "Christian Rap", "Pop/Funk", "Jungle", "Native American", "Cabaret",
    "New Wave", "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi",
    "Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical",
    "Rock & Roll", "Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing",
    "Fast-Fusion", "Bebob", "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde",
    "Gothic Rock", "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band",
    "Chorus", "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson",
    "Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass", "Primus",
    "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
    "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet",
    "Punk Rock", "Drum Solo", "A capella", "Euro-House", "Dance Hall",
    "Goa", "Drum & Bass", "Club House", "Hardcore", "Terror",
    "Indie", "BritPop", "NegerPunk", "Polsk Punk", "Beat",
    "Christian Gangsta", "Heavy Metal", "Black Metal", "Crossover", "Contemporary C",
    "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
    "SynthPop"
};

int GenreNumToString (char* GenreStr, const int GenreNum) {
    if ( GenreNum > 0 && GenreNum < (int)sizeof(GenreList) / (int)sizeof(*GenreList) ) {
        strcpy ( GenreStr, GenreList[GenreNum-1] );
        return 0;
    } else {
        strcpy(GenreStr, "unknown");
        return 1;
    }
}

int GenreStringtoNum(char *gnre) {
	int i;
	
    for(i = 0; i < (int)sizeof(GenreList) / (int)sizeof(*GenreList); i++) {
        if(_stricmp(GenreList[i], gnre) == 0) {
            return i+1;
        }
    }
    return 255;
}

int strpos2(const char *s, const char c, int n) {
	int i;

	for(i = 0; i < n; i++) {
		if(s[i] == c)
			return i;
   }
   return -1;
}

int strstr2(const char *s1, const char *s2, int ns1) {
	int i, ns2;
	
	ns2 = strlen(s2);
	for(i = 0; i < ns1; i++) {
		if(strncmp(s1+i, s2, ns2) == 0)
			return i;
	}
	return 0;
}
	
unsigned int OneByte(const char *data) {
	unsigned int n;

	n =  ((unsigned int)data[0] & 0xff) << 24;
	n += ((unsigned int)data[1] & 0xff) << 16;
	n += ((unsigned int)data[2] & 0xff) << 8;
	n += ((unsigned int)data[3] & 0xff);
  return n;
}

char *TimeString(int seconds) {
	int mins, secs;
	static char s[20];
	
	mins = (seconds / 60) % 60;
	secs = seconds % 60;
	sprintf(s, "%d:%02d", mins, secs);
	return s;
}

void ParseTag(char *key, char *data, int size, Song *tag) {
	char genre[1024];

	data[size] = 0;
	size++;
	if(strncmp(key, "nam", 4) == 0) {
      strncpy(tag->title, data, size);
	}
 	else if(strncmp(key, "ART", 4) == 0) {
      strncpy(tag->artist, data, size);
	}
 	else if(strncmp(key, "alb", 4) == 0) {
      strncpy(tag->album, data, size);
	}
 	else if(strncmp(key, "cmt", 4) == 0) {
      strncpy(tag->comment, data, size);
	}
 	else if(strncmp(key, "day", 4) == 0) {
      strncpy(tag->year, data, size);
	}
 	else if(strncmp(key, "too", 4) == 0) {
      strncpy(tag->tool, data, size);
	}
 	else if(strncmp(key, "gnre", 4) == 0)  {
		GenreNumToString(genre, data[1] );
		strcpy(tag->genrename, genre);
      tag->genrenum = data[1];
	}
 	else if(strncmp(key, "gen", 4) == 0) {
      strncpy(tag->cgen, data, size);
	}
 	else if(strncmp(key, "trkn", 4) == 0)  {
		tag->track = data[3];
	}
 	else if(strncmp(key, "disk", 4) == 0) {
		tag->disk = data[3];
	}
 	else if(strncmp(key, "wrt", 4) == 0) {
		strncpy(tag->writer, data, size);
	}
 	else if(strncmp(key, "cpil", 4) == 0) {
		tag->cpil = data[0];
	}
 	else if(strncmp(key, "tmpo", 4) == 0) {
		tag->tmpo = data[1] + (data[0]<<8);
	}
} // parsetag

void PrintTag(Song *tag) {
	int bitrate;

	printf("Format:  %s\n", tag->codec);
	printf("Details: ");
	printf("%d Hz ", tag->sample_rate);
	if(tag->channels == 2)
		printf("stereo, ");
	else if (tag->channels == 1)
		printf("mono, ");
	else
		printf("%d, ", tag->channels);
		
   if(tag->playtime != 0) {		
		bitrate = ((tag->mdatsize * 8) / tag->playtime) / 1000;
		printf("%d kbps, ", bitrate);
	}
   printf("playtime %s\n", TimeString(tag->playtime) );
	
	printf("Artist:  %s\n", tag->artist);
	printf("Album:   %s\n", tag->album );
	printf("Title:   %s\n", tag->title );
	if(tag->track == 0)
		printf("Track:\n");
	else		
		printf("Track:   %d\n", tag->track);
	if(tag->cgen[0] != 0)
		printf("Genre2:  %s\n", tag->cgen);	
	else
		printf("Genre:   %s\n", tag->genrename);
	printf("Year:    %s\n", tag->year);
	if(tag->comment[0] != 0)
		printf("Comment: %s\n", tag->comment);
	if(tag->tool[0] != 0)
		printf("Tool:    %s\n", tag->tool);		
}
			
void PrintAtom(Atom *atom){
	printf("%*s%s: %d bytes at %d\n", 2*(atom->level-1), "", atom->name, atom->size, atom->pos);
}
	
void stsd(Song *tag, char *atomdata) {
	int num, stsdOff, i, size;
	char data_format[5], data[1024], encoder_vendor[1024];
	
	 num = OneByte(atomdata+4);
    stsdOff = 8;
    for (i = 0; i < num; i++) {
    	size = OneByte(atomdata+stsdOff);
    	stsdOff += 4;
      memcpy(data_format, atomdata+stsdOff, 4);
      data_format[4] = 0;
    	stsdOff += 12;
    	memcpy(data, atomdata+stsdOff, (size - 4 - 12));
    	stsdOff += (size - 4 - 12);

    	memcpy(encoder_vendor, data+4, 4);
    	if (encoder_vendor[0] == 0) {
	    	tag->sample_rate = ((data[16]&0xff) << 8) + (data[17]&0xff);
     		tag->channels = (data[8] << 8) + data[9];
            if(strcmp(data_format, "mp4a") == 0)
	        		strcpy(tag->codec, "AAC");
            else
              strncpy(tag->codec, data_format, sizeof(tag->codec));
    	} // if
    } // for

} // stsd
				
int ParseAtom(FILE *f, unsigned int firstOff, unsigned int stopAt, FileInfo *fileinfo) {
	int types, i, NextTagPosition, time_scale, duration;
	unsigned int atomsize, end, size, off;
	char AtomHeader[10], atomname[5], atomdata[20000];
	char mRaw[1024], mKey[1024], mData[1024];
	Atom atom;
	static int level = 0;

    ++level;

    fseek(f, 0, SEEK_END);
    end = ftell(f);

    off = firstOff;
    while(off < stopAt) {
        fseek(f, off, SEEK_SET);

        fread(AtomHeader, 1, 8, f);
		  atomsize = OneByte(AtomHeader);
        memcpy(atomname, AtomHeader+4, 4);
        atomname[4] = 0;

//			dprintf("%s (%u bytes at %u)\n", atomname, atomsize, off);

        if((off + atomsize) > end) {
            printf("atom at %d claims %d bytes, end = %d\n", off, atomsize, end);
            return 0;
        }

		memcpy(atom.name, atomname, 5);
		atom.size = atomsize;
		atom.pos = ftell(f)-8;
		atom.level = level;

      if( (fileinfo->atoms = (Atom **)realloc(fileinfo->atoms,
      	(fileinfo->entries+1)*sizeof(*fileinfo->atoms))) == NULL) {
	      	printf("atomparser: couldn't reallocate memory for entry %d\n", fileinfo->entries);
	      	return 0;
	   }
		if( (fileinfo->atoms[fileinfo->entries] = (Atom *)malloc(sizeof(atom))) == NULL) {
			printf("atomparser: couldn't allocate memory for entry %d\n", fileinfo->entries);
			return 0;
		}
  		memcpy(fileinfo->atoms[fileinfo->entries++], &atom, sizeof(atom));

		if(strcmp(atomname, "mdat") == 0)
			fileinfo->tag.mdatsize = atomsize;
		else if(strcmp(atomname, "moov") == 0)
			fileinfo->moovpos = ftell(f) - 8;
		else if(strcmp(atomname, "udta") == 0) {
			fileinfo->udtapos  = ftell(f) - 8;
			fileinfo->udtasize = atomsize;
		}
		else if(strcmp(atomname, "free") == 0) {  	// we'd like a free under a moov
			if(atom.level == 2) { 							// so check for level = 1
				for(i = fileinfo->entries-1; i > 0; i--)	// go backwards through the entries
					if(fileinfo->atoms[i]->level == 1)		// until we hit a level 1
						if(strncmp(fileinfo->atoms[i]->name, "moov", 4) == 0) {
							if(atomsize > fileinfo->freesize) {	// is this free bigger than others?
								fileinfo->freepos = ftell(f) - 8;
								fileinfo->freesize = atomsize;						
							}
							else
                        break;
                  } // if moov
			} // if level = 2
		} // if free

//if it's a container atom, parse the contents of the atom
		  types = (int)sizeof (AtomType) / (int)sizeof (*AtomType);
		  for(i = 0; i < types; i++) {
	        	if(strcmp(atomname, AtomType[i]) == 0) {        	
   	         ParseAtom(f, ftell(f), off+atomsize, fileinfo);
      	      break;
      	   }
        }

// meta atom contains tags and some other data
        if(strcmp(atomname, "meta") == 0) {
	         fread(atomdata, 1, atomsize-4, f); // read in meta atom
            NextTagPosition = strpos2(atomdata, '', atomsize-4);  // find the start of first tag
				NextTagPosition = strstr2(atomdata, "ilst", atomsize-4) + 8;
            while ((unsigned int)NextTagPosition < (atomsize-4) && NextTagPosition > 8) {
					 size = OneByte(atomdata+(NextTagPosition-4)) - 4;
					if(size >= 1024)
						break;
 					 memcpy(mRaw, atomdata+NextTagPosition, size);
                memcpy(mKey, mRaw, 4);
                memcpy(mData, mRaw+20, size-20);
                NextTagPosition += size + 4;

                ParseTag(mKey, mData, size-20, &(fileinfo->tag));
            } // while
        } // if meta

// mdhd has data for calculating playtime
        if(strcmp(atomname, "mdhd") == 0) {
            fread(atomdata, 1, 20, f);
 			   time_scale = OneByte(atomdata+12);
			   duration   = OneByte(atomdata+16);
			   fileinfo->tag.playtime = duration / time_scale;
        } // if mdhd

// stsd has data for sample, channels and codec
        if(strcmp(atomname, "stsd") == 0) {
            fread(atomdata, 1, atomsize-8, f);
            stsd(&(fileinfo->tag), atomdata);
        }

// bump offest and continue
        if(atomsize == 0) {  // a 0 byte atom is the end
            off = stopAt;
        }
        else {
            off += atomsize;
        }

    } // while

    --level;
    return 1;
} // parseatom

void FreeFileInfo(FileInfo *fileinfo) {
	int i;

	for(i = 0; i < fileinfo->entries; i++)
		free(fileinfo->atoms[i]);
	free(fileinfo->atoms);
	
	fileinfo->entries = 0;
	fileinfo->atoms = NULL;
}		

/*----------------------------------------------
int main(int argc, char *argv[]) {

   FILE *f;
   int end, i, debug = 0;
   FileInfo fi;
   char infile[1024];

	if(argc > 1)
		strcpy(infile, argv[1]);
	else {
		printf("no file specified\n");
		exit(1);
	}
			
   if(argc > 2)
      if(strncmp(argv[2], "--debug", 8) == 0)
      	debug = 1;

   fi.entries = 0;

	if( (f = fopen(infile, "rb")) == NULL) {
		printf("couldn't open %s\n", infile);
		exit(1);
	}
	
   fseek(f, 0, SEEK_END);
   end = ftell(f);

   ParseAtom(f, 0, end, &fi);

   if(debug) {
      for(i = 0; i < fi.entries; i++) {
      	PrintAtom(fi.atoms[i]);
      }
   }

   printf("\n%s\n", argv[1]);
   PrintTag(&fi.tag);

	fclose(f);
	
}
-------------------------------------------*/
