/* 
 * tg- a command line tagger for mp4 and related files				(c) Richard 2003
 */

#define VERSION "Version 0.36  2003-08-22"

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utime.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <process.h>
#include <io.h>

#include "tg.h"
#include "atomparser.h"

#define DEBUG 1
#define TAGSIZE 750

// this prevents command line wildcard expansion
// so that we can do it ourselves
int _CRT_glob = 0;

void store4int(unsigned int n, char *data) {
  data[0] = (n >> 24) & 0xff;
  data[1] = (n >> 16) & 0xff;
  data[2] = (n >>  8) & 0xff;
  data[3] = (n >>  0) & 0xff;
}

void dprintf (const char *fmt, ...)
{
	va_list ap;

   if(DEBUG) {	
      va_start (ap, fmt);
      vprintf (fmt, ap);
      va_end (ap);
   }
}

	
int CreateUDTA(Song *tag, char *fulltag, int len) {

	static char udta[] = {0,0,0,0,'u','d','t','a'};
	static char meta[] = {0,0,0,0,'m','e','t','a',0,0,0,0};
	static char hdlr[] = {0,0,0,0,'h','d','l','r',0,0,0,0,0,0,0,0,'m','d','i','r',
		'a','p','p','l',0,0,0,0,0,0,0,0,0,0};
	static char ilst[] = {0,0,0,0,'i','l','s','t'};		
	static char cnam[] = {0,0,0,0,'','n','a','m',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char cart[] = {0,0,0,0,'','A','R','T',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char calb[] = {0,0,0,0,'','a','l','b',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};		
	static char gnre[] = {0,0,0,0,'g','n','r','e',0,0,0,0,'d','a','t','a',0,0,0,0,0,0,0,0,0,0};
	static char cgen[] = {0,0,0,0,'','g','e','n',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char trkn[] = {0,0,0,0,'t','r','k','n',0,0,0,0,'d','a','t','a',0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0};
	static char disk[] = {0,0,0,0,'d','i','s','k',0,0,0,0,'d','a','t','a',0,0,0,0,0,0,0,0,0,0,
		0,1,0,1};
	static char cday[] = {0,0,0,0,'','d','a','y',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char cpil[] = {0,0,0,0,'c','p','i','l',0,0,0,0,'d','a','t','a',0,0,0,21,0,0,0,0,0};	
	static char tmpo[] = {0,0,0,0,'t','m','p','o',0,0,0,0,'d','a','t','a',0,0,0,21,0,0,0,0,0,0};		
	static char ctoo[] = {0,0,0,0,'','t','o','o',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	static char ccmt[] = {0,0,0,0,'','c','m','t',0,0,0,0,'d','a','t','a',0,0,0,1,0,0,0,0};
	int pos;

// to do: make sure we don't write beyond end of fulltag

// append each atom to fulltag
// pos serves as a running total of the size of the atom
	pos = 0;
	memcpy(fulltag, udta, sizeof(udta));
	pos += sizeof(udta);
	memcpy(fulltag+pos, meta, sizeof(meta));
	pos += sizeof(meta);
	memcpy(fulltag+pos, hdlr, sizeof(hdlr));
	pos += sizeof(hdlr);
	memcpy(fulltag+pos, ilst, sizeof(ilst));
	pos += sizeof(ilst);
	
	// song title (string)
	if(tag->title[0] != 0) {
   	store4int(sizeof(cnam)+strlen(tag->title), cnam);
   	store4int(sizeof(cnam)+strlen(tag->title)-8, cnam+8);
   	memcpy(fulltag+pos, cnam, sizeof(cnam));
   	pos += sizeof(cnam);
   	memcpy(fulltag+pos, tag->title, strlen(tag->title));
   	pos += strlen(tag->title);
	}
	
	// artist name (string)
	if(tag->artist[0] != 0) {
   	store4int(sizeof(cart)+strlen(tag->artist), cart);
   	store4int(sizeof(cart)+strlen(tag->artist)-8, cart+8);
   	memcpy(fulltag+pos, cart, sizeof(cart));
   	pos += sizeof(cart);
   	memcpy(fulltag+pos, tag->artist, strlen(tag->artist));
   	pos += strlen(tag->artist);
	}
	
	// album name (string)
	if(tag->album[0] != 0) {
   	store4int(sizeof(calb)+strlen(tag->album), calb);
   	store4int(sizeof(calb)+strlen(tag->album)-8, calb+8);
   	memcpy(fulltag+pos, calb, sizeof(calb));
   	pos += sizeof(calb);
   	memcpy(fulltag+pos, tag->album, strlen(tag->album));
   	pos += strlen(tag->album);
	}
	
	// genre number (id3 + 1)
	if(tag->genrenum != 0 && tag->genrenum != 255) {
   	store4int(sizeof(gnre), gnre);
   	store4int(sizeof(gnre)-8, gnre+8);
   	gnre[sizeof(gnre)-1] = tag->genrenum;
   	memcpy(fulltag+pos, gnre, sizeof(gnre));
   	pos += sizeof(gnre);
	}
 	else if(tag->cgen[0] != 0) {  	// genre name (string)
   	store4int(sizeof(cgen)+strlen(tag->cgen), cgen);
   	store4int(sizeof(cgen)+strlen(tag->cgen)-8, cgen+8);
   	memcpy(fulltag+pos, cgen, sizeof(cgen));
   	pos += sizeof(cgen);
   	memcpy(fulltag+pos, tag->cgen, strlen(tag->cgen));
   	pos += strlen(tag->cgen);
	}

	// track number & total tracks
	if(tag->track != 0) {
   	store4int(sizeof(trkn), trkn);
   	store4int(sizeof(trkn)-8, trkn+8);	
   	trkn[sizeof(trkn)-5] = tag->track;
   	trkn[sizeof(trkn)-3] = tag->track;
   	memcpy(fulltag+pos, trkn, sizeof(trkn));
   	pos += sizeof(trkn);
	}
	
	// disk number & total disks
	if(0) {
   	store4int(sizeof(disk), disk);
   	store4int(sizeof(disk)-8, disk+8);	
   	disk[sizeof(disk) - 3] = 1;
   	disk[sizeof(disk) - 1] = 1;
   	memcpy(fulltag+pos, disk, sizeof(disk));
   	pos += sizeof(disk);
	}
	
	// year (string)
	if(tag->year[0] != 0) {	
   	store4int(sizeof(cday)+strlen(tag->year), cday);
   	store4int(sizeof(cday)+strlen(tag->year)-8, cday+8);
   	memcpy(fulltag+pos, cday, sizeof(cday));
   	pos += sizeof(cday);strlen(tag->year);
   	memcpy(fulltag+pos, tag->year, strlen(tag->year));
   	pos += strlen(tag->year);
	}
	
	// part of a compilation (0 or 1)
	if(0) {
   	store4int(sizeof(cpil), cpil);
   	store4int(sizeof(cpil)-8, cpil+8);	
   	memcpy(fulltag+pos, cpil, sizeof(cpil));
   	pos += sizeof(cpil);
	}
	
	// tempo in BPM
	if(0) {
   	store4int(sizeof(tmpo), tmpo);
   	store4int(sizeof(tmpo)-8, tmpo+8);	
   	memcpy(fulltag+pos, tmpo, sizeof(tmpo));
   	pos += sizeof(tmpo);
	}   	

	// tool used to create file (string)
	if(tag->tool[0] != 0) {
   	store4int(sizeof(ctoo)+strlen(tag->tool), ctoo);
   	store4int(sizeof(ctoo)+strlen(tag->tool)-8, ctoo+8);	
   	memcpy(fulltag+pos, ctoo, sizeof(ctoo));
   	pos += sizeof(ctoo);
   	memcpy(fulltag+pos, tag->tool, strlen(tag->tool));
   	pos += strlen(tag->tool);
	}

	// comment (string)
	if(tag->comment[0] != 0) {
   	store4int(sizeof(ccmt)+strlen(tag->comment), ccmt);
   	store4int(sizeof(ccmt)+strlen(tag->comment)-8, ccmt+8);
   	memcpy(fulltag+pos, ccmt, sizeof(ccmt));
   	pos += sizeof(ccmt);
   	memcpy(fulltag+pos, tag->comment, strlen(tag->comment));
   	pos += strlen(tag->comment);
	}

// done with adding tags
// now clean up the atomsizes and perhaps add a free atom
// the free atom gives us some space if we edit tags later
	
	store4int(pos-20-sizeof(hdlr), fulltag+20+sizeof(hdlr)); // size of ilst
	
	if(pos < (len + 8) ) {
		store4int((len-pos), fulltag+pos);
		memcpy(fulltag+pos+4, "free", 4);
		pos = len;
	}
	
	store4int(pos, fulltag);		// size of entire atom at beginning
	store4int(pos-8, fulltag+8);	// size of meta at meta position
	store4int(sizeof(hdlr), fulltag+20);  // size of hdlr at it position

	return pos;
} // createudta

// take the moov atom and copy it to end
// change original moov atom to free atom
// change atomsize of new moov atom to accomodate new tags
int MOOVtoEnd(FILE *f, int len, int moovpos) {
	char AtomHeader[10], *moov;
	int atomsize;

   if(moovpos == 0)
   	return 0;

	fseek(f, moovpos, SEEK_SET);

   fread(AtomHeader, 1, 8, f);
   atomsize = OneByte(AtomHeader);

   fseek(f, moovpos, SEEK_SET);  	// move to start of moov atom
	if( (moov = (char *)malloc(atomsize)) == NULL) {
		printf("MOOVtoEND: memory allocation failed\n");
  		exit(1);
  	}
	fread(moov, 1, atomsize, f);   	// read entire moov atom
	
   fseek(f, moovpos+4, SEEK_SET); // back to start of atom
   fwrite("free", 1, 4, f);       // change old moov atom to free

   fseek(f, 0, SEEK_END);         	// moov atom to end of file
   store4int(atomsize+len, moov);	// bump the len of the moov
   fwrite(moov, 1, atomsize, f);   // write the new moov atom

   free(moov);
   return 1;
} //moovtoend


// assumes path is in form  artist\album\nn - title.ext
// or artist\album\something - nn - title.ext
void parsepath(char *path, Song *tag) {
	char dir[_MAX_DIR], name[_MAX_FNAME], *s;
	int r;

	_splitpath(path, NULL, dir, name, NULL);

	tag->track = atoi(name);
	if(tag->track >= 1 && tag->track <= 99)
		strcpy(tag->title, name+5);
	else {
		if( (s = strchr(name, '-')) != NULL) {
			tag->track = atoi(s+2);
			strncpy(tag->title, s+7, strlen(s)-7);
		}
 	}

	r = strlen(dir);
	dir[r-1] = 0;
	--r;
	while(--r)
		if(dir[r] == '/' || dir[r] == '\\') {
			strcpy(tag->album, dir+r+1);
			dir[r] = 0;
			break;
		}
		
	while(--r >= 0)
		if(dir[r] == '/' || dir[r] == '\\') {
			strcpy(tag->artist, dir+r+1);
			break;
		}

} // parsepath

void usage(void) {
printf("Tagger for mp4's - uses iTunes format.  Tested only with QT created mp4's.\n");
printf("** Use at your own risk **                    (c) Richard 2003\n");
printf(VERSION"\n");
printf("\n");
printf("Usage: tg <filename> Options\n");
printf("Wildcards are supported in filename [more or less]\n");
printf("\n");
printf("Options:\n");
printf("[none]      print existing tags\n");
printf("--debug     print existing tags and atom info\n");
printf("--remove    remove existing tags\n");
printf("--auto      parses filepath - must be in form artist\\album\\nn - title.ext\n");
printf("--artist  <value>\n");
printf("--album   <value>\n");
printf("--title   <value>\n");
printf("--track   <value>\n");
printf("--genre   <value>\n");
printf("--year    <value>\n");
printf("--comment <value>\n");
printf("--tool    <value>  [tools used to create the file]\n");
printf("\n");
printf("use mp4creator.exe to optimize mp4s\n");
printf("thanks to www.getid3.org and tag for inspiration");

} // usage

int tagger(int argc, char *filename, char *argv[]) {
	Song tag;
	FileInfo fi;
	char infile[MAX_PATH+1], outfile[MAX_PATH+1], fulltag[2048];
   char str[1024];
	char *buf;
	int end, changetags, i, len, wrote;
	int autoo, back, debug, removetags;
	FILE *inf, *outf;
	struct stat tim;
	struct utimbuf ut;

// clear these structures
	strncpy((char *)&tag, "", sizeof(tag));
	strncpy((char *)&fi, "", sizeof(fi));

	_fullpath(infile, filename, sizeof(infile)-1);
	stat(infile, &tim);	// remember file times
	ut.actime = tim.st_atime;
	ut.modtime = tim.st_mtime;

	if( (inf  = fopen(infile, "rb")) == NULL) {
		printf("File not found: %s\n", infile);
		return 0;
	}
	
	fseek(inf, 0, SEEK_END);
	end = ftell(inf);
	fseek(inf, 0, SEEK_SET);

// read the existing tags
   if(ParseAtom(inf, 0, end, &fi) == 0) {
      printf("problem reading atoms in %s\n", infile);
      return 0;
   }
	memcpy(&tag, &(fi.tag), sizeof(Song));
	
// if auto, read tag fields from pathname.
// cmd line tags override pathname, which is why this is before general cmd line read
    changetags = 0;
    autoo = 0;
    for(i = 1; i < argc; i++) {
        if(strcmp(argv[i], "--auto") == 0) {
	         autoo = 1;
            changetags = 2;
            parsepath(infile, &tag);
            break;
        }
    }

// read params from command line
    back = 0;
    removetags = 0;
    debug = 0;
    for(i = 1; i < argc; i++) {
	    if(strcmp(argv[i], "--artist") == 0) {
				strncpy(tag.artist, argv[++i], sizeof(tag.artist)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--album") == 0){
				strncpy(tag.album, argv[++i], sizeof(tag.album)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--title") == 0) {
				strncpy(tag.title, argv[++i], sizeof(tag.title)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--comment") == 0) {
				strncpy(tag.comment, argv[++i], sizeof(tag.comment)-1);
            changetags = 1;
        }
        else if(strcmp(argv[i], "--track") == 0) {
	        tag.track = atoi(argv[++i]);
	        changetags = 1;
	     }
        else if(strcmp(argv[i], "--genre") == 0) {
  				strncpy(tag.cgen, argv[++i],sizeof(tag.cgen)-1);
				strncpy(tag.genrename, argv[i],sizeof(tag.genrename)-1);
				tag.genrenum = GenreStringtoNum(argv[i]);
				if(tag.genrenum == 255) {
					printf("Warning: unrecognized genre %s\n", argv[i]);
				}
            changetags = 1;
        }
        else if(strcmp(argv[i], "--year") == 0) {
				strncpy(tag.year, argv[++i], sizeof(tag.year)-1);		
            changetags = 1;
        }
        else if(strcmp(argv[i], "--tool") == 0) {
				strncpy(tag.tool, argv[++i], sizeof(tag.tool)-1);		
            changetags = 1;
        }
        else if(strcmp(argv[i], "--back") == 0) {
	        back = 1;
	     }
        else if(strcmp(argv[i], "--debug") == 0) {
	        debug = 1;
	     }
        else if(strcmp(argv[i], "--remove") == 0) {
	        removetags = 1;
	        changetags = 1;
	     }
	     else if( strcmp(argv[i], "--auto") == 0) {
		     autoo = 1;
		  }
	     else {
            printf("unknown field: %s\n", argv[i]);
            FreeFileInfo(&fi);
            return 0;
        }  // if chain
    }  // for each argv

// if we are not changing tags, just report tag info and return
	if(changetags == 0) {
   	if(debug) {
      	for(i = 0; i < fi.entries; i++)
	      	PrintAtom(fi.atoms[i]);
      } // if debug
      printf("\n%s\n", infile);
		if(fi.udtapos != 0)
            PrintTag(&fi.tag);
        else
            printf("no tags\n");
        FreeFileInfo(&fi);
        return 1;
    } // no change

// we are changing something, so create an outfile
	sprintf(outfile, "%s.bak", infile);
	if( (outf = fopen(outfile, "w+b")) == NULL) {
		printf("Can't open backup file %s\n", outfile);
		FreeFileInfo(&fi);
		return 0;
	}

//copy inf to outfile
	if( (buf = (char *)malloc(end)) == NULL) {
		printf("tagger: memory allocation failure\n");
		exit(1);
	}
	fseek(inf, 0, SEEK_SET);
	fread(buf, 1, end, inf);
	fwrite(buf, 1, end, outf);
	free(buf);
	fclose(inf);
	
//if remove, do it
// we remove tags by changing the udta atom with the tags to a free atom
    if(removetags == 1) {
        if(fi.udtapos == 0)
            printf("no existing tags\n");
        else {
				for(i = 0; i < fi.entries; i++) {
					if(strcmp(fi.atoms[i]->name, "udta") == 0) {
						fseek(outf, fi.atoms[i]->pos+4, SEEK_SET);
						fwrite("free", 1, 4, outf);
					}					
				}
				if(back == 0) {			// rename backup file to real name
					remove(infile);
					fclose(outf);
					utime(outfile, &ut); // change file times to match original
					rename(outfile, infile);
				}
				printf("tags removed from %s\n", infile);
        } // else
        FreeFileInfo(&fi);
        return 1;
    } // if remove

// this section creates the new tag

	len = CreateUDTA(&tag, fulltag, 0);	

// is there an existing udta atom we can overwrite?
// if there's enough room in the atom, overwrite
// if not, change to a free atom
   wrote = 0;
   if(fi.udtapos != 0) {
		if(((unsigned int)len < fi.udtasize) && (wrote == 0)) {
      	CreateUDTA(&tag, fulltag, fi.udtasize);
         fseek(outf, fi.udtapos, SEEK_SET);
         fwrite(fulltag, 1, fi.udtasize, outf);
         wrote = 1;
       } // if
       else {
       	fseek(outf, fi.udtapos+4, SEEK_SET);
         fwrite("free", 1, 4, outf);
       }
   } // there's a udta

// is there an existing free atom we can overwrite?
// if there's enough room in the atom, overwrite
   if(fi.freepos != 0 && wrote == 0) {
		if(((unsigned int)len < fi.freesize) && (wrote == 0)) {
      	CreateUDTA(&tag, fulltag, fi.freesize);
         fseek(outf, fi.freepos, SEEK_SET);
         fwrite(fulltag, 1, fi.freesize, outf);
         wrote = 1;
       } // if
   } // there's a free

// if we didn't overwite an existing udta or free, add to a moov
	if(wrote == 0) {
		len = CreateUDTA(&tag, fulltag, TAGSIZE);	
   	MOOVtoEnd(outf, len, fi.moovpos);
   	fwrite(fulltag, 1, len, outf);
   }

// done.  close the file
	fclose(outf);
	fclose(inf);
	
// unless we want a backup, rename temp file to original name and delete original  	
	if(back == 0) {
		remove(infile);
		rename(outfile, infile);
		strcpy(outfile, infile);
	}

// if we used moovtoend, there is a large free atom.  use mpcreator to eliminate it
// if mp4creator60 is not present, nothing happens
// consider using mpeg4ip's optimize - same source as mp4creator
	if(wrote == 0) {
		sprintf(str, "\"%s\"", outfile);
		spawnlp(P_WAIT, "mp4creator60.exe", "mp4creator60.exe",
			"-optimize", str, NULL);
	}


// reopen the resulting file to read its tags
	strncpy((char *)&fi, "", sizeof(fi));	
  	outf = fopen(outfile, "rb");
	fseek(outf, 0, SEEK_END);
	end = ftell(outf);
	ParseAtom(outf, 0, end, &fi);
	printf("\n%s\n", outfile);
   PrintTag(&fi.tag);
   fclose(outf);

// change file times to match original
  	utime(outfile, &ut);
			
	FreeFileInfo(&fi);
   return 1;	
}	// tagger

int pstrcmp(const void *p1, const void *p2) {
	return strcmp(*(char * const *)p1, *(char * const *)p2);
}


void AddString(StringList *sl, char *s) {

	if( (sl->item = (char **)realloc(sl->item,
      	(sl->entries+1)*sizeof(*sl->item))) == NULL) {
		printf("AddString: couldn't reallocate memory for entry %d\n", sl->entries);
	   exit(1);
 	}

	if( (sl->item[sl->entries] = (char *)malloc(strlen(s)+1)) == NULL) {
		printf("AddString: couldn't allocate memory for entry %d\n", sl->entries);
		exit(1);
	}

	strcpy(sl->item[sl->entries++], s);
} // AddString

void FreeStringList(StringList *sl) {
	int i;

	for(i = 0; i < sl->entries; i++)
		free(sl->item[i]);
	free(sl->item);
	
	sl->entries = 0;
	sl->item = NULL;
}

// some rudimentary wildcard processing
// given a pathmask, produce a matching list of files
// does not recurse into subdirs
void dirlist(StringList *filelist, char *pathmask) {
	char s[MAX_PATH+1];
	char drive[_MAX_DRIVE+1], dir[_MAX_DIR+1], name[_MAX_FNAME+1], ext[_MAX_EXT+1];
	struct _finddata_t fd;
	int fhand;

	_splitpath(pathmask, drive, dir, name, ext);
// if no name given, add * at end
	if(*name == 0) {
		sprintf(s, "%s%s*", drive, dir);
	}
	else {
		fhand = _findfirst(pathmask, &fd);
// if the path is a subdir and name does not include *, add * to end
		if(fd.attrib & _A_SUBDIR && strchr(name, '*') == 0) {
				strcat(dir, name);
				strcat(dir, "\\");
				sprintf(s, "%s%s*", drive, dir);
		}
		else {
			strncpy(s, pathmask, MAX_PATH);
		}
		_findclose(fhand);
	}

// use findfirst/findnext to create filelist - a list of all files
// 	matching our pattern

	if ( (fhand = _findfirst(s, &fd)) != -1) {
		do {
			if((fd.attrib & _A_SUBDIR) == 0) {
				sprintf(s, "%s%s%s", drive, dir, fd.name);
				AddString(filelist, s);
			}
		} while(_findnext(fhand, &fd) != -1);
		_findclose(fhand);
		qsort((void *)(filelist->item), filelist->entries, sizeof(*filelist->item), pstrcmp);
	}
} // dirlist

int main(int argc, char *argv[]) {
	int i;
	StringList filelist = {0, NULL};

	if(argc == 1) {
		usage();
		exit(1);
	}

// get a list of filenames matching argv[1]
	dirlist(&filelist, argv[1]);
	
// tag each file in the filelist
	for(i = 0; i < filelist.entries; i++)
		tagger(argc-1, filelist.item[i], argv+1);

// not necessary at the end of the program
	FreeStringList(&filelist);
	
	return 1;
}

